###########################################################################
# Author List
##############
# Jaeyeon Jung - MIT CSAIL
# Stuart Schechter - Harvard University DEAS
# Will Stockwell - MIT CSAIL
#
# Copyright (c) 2004 Jaeyeon Jung, Stuart Schechter, Will Stockwell
#                    Cambridge, MA, All rights reserved
#
#######################################################################
# Documentation
################
#
# Subroutines module used in data collection process for the SSH project
# of the NMS group at MIT CSAIL.  Pleaes visit the following URL for more
# information:
#
# http://nms.lcs.mit.edu/projects/ssh/collect-ssh.php
#
# Module subroutines:
#
#  Print_Result
#  Anonymize_IP
#  Read_Host
#  Read_Known_Hosts
#  Read_Authorized_Keys
#  Read_Public_Keys
#  Ignore_User
#  Encrypt_Results
#  Save_Results
#  Submit_Results_Mail
#  Submit_Results_HTTP
#  Homedir_Netfs_IP
#  Read_User
#
# Exposed module variables:
# 
# $verbose
# @ignore_users
# $ifconfig
# 
#
#######################
# Terms and conditions
#######################
# This script is to be integrated into the OpenSSL package and licensed
# under the same terms specified for the rest of the OpenSSH package,
# as copied below.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
###########################################################################

package CollectSSH;

use strict;
use Socket qw(:DEFAULT :crlf);

# Modules included in the collect-ssh tarball just in case they're not
# installed on the system.  Please see README for information about why
# these algorithms are necessary
use Crypt::Rijndael;
use Digest::SHA1;

use vars (
          qw(@ISA @EXPORT_OK $VERSION), 
          qw(&Encrypt_Results &Save_Results &Submit_Results_Mail),
          qw(&Submit_Results_HTTP &Read_User &Read_Host &Ignore_User)
);

require Exporter;
$VERSION = '1.00';
@ISA = ('Exporter');

@EXPORT_OK = (
              qw(&Encrypt_Results &Save_Results &Submit_Results_Mail),
              qw(&Submit_Results_HTTP &Read_User &Read_Host &Ignore_User),
              qw($VERBOSE $DEBUG @IGNORE_USERS)
);

# Locate the ifconfig command in the filesystem
my @ifconfig_paths = ("/sbin/ifconfig", 
                       "/usr/sbin/ifconfig"
                       );
our $ifconfig = "";

foreach (@ifconfig_paths) {
    if (-e $_) {
        $ifconfig = $_;
        last;
    }
}

# Used in Print_Result to determine whether to print results as they are
# recorded
our $VERBOSE = 1;
our $DEBUG = 0;

# Used in Ignore_User to determine whether to in fact ignore a particular user
# This list was built looking at system users on Fedora Core 2, SunOS 5.9,
# and OpenBSD 3.5.  It is likely incomplete. User exclusion parameters can
# be implemented by adding user names to this list

our @IGNORE_USERS = (
                     "daemon",
                     "operator",
                     "bin",
                     "smmsp",
                     "popa3d",
                     "adm",
                     "lp",
                     "sync",
                     "shutdown",
                     "halt",
                     "mail",
                     "news",
                     "uucp",
                     "operator",
                     "games",
                     "gopher",
                     "ftp",
                     "sshd",
                     "rpm",
                     "vcsa",
                     "nscd",
                     "sshd",
                     "rpc",
                     "rpcuser",
                     "nfsnobody",
                     "pcap",
                     "mailnull",
                     "smmsp",
                     "dbus",
                     "xfs",
                     "ntp",
                     "gdm",
                     "_portmap",
                     "_identd",
                     "_rstatd",
                     "_rusersd",
                     "_fingerd",
                     "_x11",
                     "_kdc",
                     "_kadmin",
                     "_spamd",
                     "uucp",
                     "www",
                     "_isakmpd",
                     "named",
                     "proxy",
                     "_syslogd",
                     "_pflogd",
                     "_bgpd",
                     "_tcpdump",
                     "nobody"
                     );

# Version of the script's output format and a variable to hold data that
# will be encrypted as results
my $COLLECT_VERSION = "20050515";
my $RESULTS = "";

# Setup Rijndael encryption context for use in IP address anonymization
my $key = pack ("H32", "6bbf3d67941ce44fd0dc413da9e1a7a8");
my $cipher = new Crypt::Rijndael $key, Crypt::Rijndael::MODE_CBC;

die "Cannot find ifconfig.  Specify \$CollectSSH:ifconfig in user program or edit \@ifconfig_paths in CollectSSH.pm to include your ifconfig's path" if ($ifconfig eq "");

#########################################################################
# Subroutine: SHA1_Hash($data)
###
#
# Uses SHA1 algorithm to hash $data.  Returns result.
#########################################################################
sub SHA1_Hash
{
    # Newline added for consistency with previous implementation
    my $h = Digest::SHA1::sha1_hex("$_[0]\n");
    chomp ($h);
    return $h;
}

#########################################################################
# Subroutine: Print_Result($string)
###
#
# Appends $string to $RESULTS variable for later submission to collection
# server. If $VERBOSE is true, also prints $string to STDOUT
#########################################################################
sub Print_Result
{
    my ($str) = @_;
    $RESULTS .= $str;
    print $str if ($VERBOSE || $DEBUG);
}

#########################################################################
# Subroutine: Anonymize_IP($ip)
###
#
# Anonymizes $ip (in dotted quad format) using scheme described in [1] 
# referenced in the README file.  Returns result in dotted quad format.
#########################################################################
sub Anonymize_IP
{
    my ($ip) = @_;
    my @ibits;
    my @obits;
    my $prefix = "";
    my ($enc, $perm);

    # Extract individual quads of the IP address
    $ip =~ /(\d{1,3})\.(\d{1,3})\.(\d{1,3})\.(\d{1,3})/
        || die "Anonymize_IP failed. ($ip) doesn't look like an IP";

    # turn ip address into an array of 32 '0's and '1's
    @ibits = split "", unpack("B32", pack ('C4', ($1, $2, $3, $4)));

    # Anonymize each bit with pseudorandom permutation of preceeding prefix
    # bits.
    foreach (0..31) {
        # Grab a 32-bit chunk of the cipher block.  All we really need is a 
        # single bit.  This works fine to avoid endianness problems
        $perm =  unpack "N", pack "H8", substr(unpack("H32", $cipher->encrypt(pack("A16", $prefix))), 24, 8);

        # Effectively an XOR of IP bit with LSB of permutation 
        $obits[$_] = ($ibits[$_] + $perm) % 2;
        $prefix .= $ibits[$_];
    }

    # Change array of 32 '0's and '1's back into a dotted quad IP
    return join (".", unpack ('C4', pack ("B32", join ("", @obits))));
}

#########################################################################
# Subroutine: Read_Host()
###
#
# Collects host-specific (anonymized IP address, netmask, OS, OS revision,
# SSH revision) data and includes it in results via Print_Result
#########################################################################
sub Read_Host
{
    my ($ip, $iphash, $mask, $os, $ssh, $time, $hostname, $ifout);

    # Grad host name from uname
    chop($hostname = `uname -n`);
    # Resolve hostname to IP address
    my ($name,$aliases,$addrtype,$length,@addrs) = gethostbyname $hostname; 

    if (! defined $addrs[0]) {
        # Resolve failed
        $iphash = "CANTRESOLVE";
    } else {
        # Anonymize local IP address
        my ($ipa, $ipb, $ipc, $ipd) = unpack('C4', $addrs[0]);
        $ip = "$ipa.$ipb.$ipc.$ipd";
        $iphash = Anonymize_IP($ip);

        # Determine associated networks
        $ifout = `$ifconfig -a`;
        # Match linux ifconfig type output
        if ($ifout =~ /inet addr:$ip(  Bcast:\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})?  Mask:(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})/) {
            $mask = $2;
        # Match SunOS & BSD type output
        } elsif ($ifout =~ /inet $ip netmask 0?x?([a-fA-F0-9]{8,8})/) {
            my ($ma, $mb, $mc, $md) = unpack "C4", pack "H8", $1;
            $mask = "$ma.$mb.$mc.$md";
        # Either no netmask available or unknown ifconfig output format
        } else {
            $mask = "NOMASK";
        }
    }

    # OS and revision information
    chop($os = `uname -s -r`);
    # SSH version (printed to STDERR, hence the redirection)
    chop($ssh = `ssh -V 2>&1`);
    # Host current time (in seconds since epoch)
    chop($time = `date +"%Y%m%d %T %Z"`);

    Print_Result "collect output $COLLECT_VERSION\n";
    Print_Result "host information\n";
    Print_Result "$iphash\n$mask\n$os\n$ssh\n$time\n";
}

#########################################################################
# Subroutine: Read_Known_Hosts($known_hosts_file)
###
#
# Collects information from known_hosts file entries (anonymized IP address,
# entry line number) and includes it in results via Print_Result
#########################################################################
sub Read_Known_Hosts
{
    my ($kh_file) = @_;
    my ($match, $host, $ip, $iphash, $khatime, $khmtime, $linenum, $pat);

    return if (! -e $kh_file);

    my @fields = stat("$kh_file");
    ($khatime, $khmtime) = ($fields[8], $fields[9]);

    Print_Result "known hosts file: $khatime $khmtime\n";

    if (!open(KH, "<$kh_file"))
    {
      print STDERR "Unable to open $kh_file: $!";
      return;
    } 

    $linenum = 1;

    # For each line in $kh_file
    foreach (<KH>) {
        $match = 0;

        # Match line against known known_hosts formats
        # Extract public key string and determine IP address either directly
        # or via DNS
        if (/,?(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}) ssh-.+ [\w\d\+\=\/]+/ ||
            /,?(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}) \d+ \d+ [\w\d\+\=\/]+/) {
            $ip = $1;
        } elsif (/([\w\d\.\-]+) ssh-.+ [\w\d\+\=\/]+/ ||
                 /([\w\d\.\-]+) \d+ \d+ [\w\d\+\=\/]+/) {
            $host = $1;
            # No IP address in these formats.  Resolve hostname
            my ($name,$aliases,$addrtype,$length,@addrs) = gethostbyname $host;

            if (! defined $addrs[0]) {
                $linenum++;
                next;
            }
            
            # Convert resolved IP to dotted quad format
            my ($ipa, $ipb, $ipc, $ipd) = unpack('C4', $addrs[0]);
            $ip = "$ipa.$ipb.$ipc.$ipd";
        } else {
            $linenum++;
            next;
        }

        # Anonymize IP address and hash public key string
        $iphash = Anonymize_IP($ip);

        Print_Result "$iphash $linenum\n";
        $linenum++;
    }
    
    close (KH);
}

#########################################################################
# Subroutine: Read_Authorized_Keys($authorized_keys_file)
###
#
# Collects information from authorized_keys file entries (SHA1-hashed public 
# key)
#########################################################################
sub Read_Authorized_Keys
{
    my ($ak_path) = @_;
    my ($pubkey, $pubkhash);

    return if (! -e $ak_path);

    if (!open (AK, "<$ak_path")) {
        print STDERR "failed to open authorized_keys file: $!";
        return;
    }

    # For each line in $ak_path, extract and hash public key string
    foreach (<AK>) {
        next if (! (/ssh-.+ ([\w\d\+\=\/]+) / || /(\d+ \d+ [\w\d\+\=\/]+)/));

        $pubkey = $1;
        $pubkhash = SHA1_Hash($pubkey);

        Print_Result "$pubkhash\n";
    }

    close (AK);
}

#########################################################################
# Subroutine: Read_Public_Keys($user_home_dir)
###
#
# Collects information for public key files in $_user_home_dir/.ssh/*.pub
# files (SHA1-hashed public key, "is associated private key encryted?" boolean)
#########################################################################
sub Read_Public_Keys
{
    my ($homedir) = @_;

    my ($privkey_file, $pubkey_file, $pkversion, $pkatime, $pkmtime,
        $pubkeyA, $pubkeyB, $pubkhash, $privisenc);
    my $ssh_path = "$homedir/.ssh";

    return if (! -e $ssh_path);

    # Find .pub files in user' ~/.ssh
    opendir SSH, $ssh_path
	or die "opendir failed for $ssh_path: $!\n";
    my @pubkeys = map {"$ssh_path/$_"} grep {/\.pub$/} readdir SSH;
    closedir SSH;

    # For each available public key file
    PKLOOP: foreach (@pubkeys) {
        $pubkey_file = $_;
        /(.+)\.pub/ || next;
        $privkey_file = $1;

        # Verify that both a public key and private key exist
        next if ((! -e "$privkey_file") || (! -e "$pubkey_file"));

        # Open _public_ key
        open (PUB, "<$pubkey_file") 
            || die "failed to open public key file: $!";
        $_ = <PUB>;
        $pubkeyA = $_;
        chomp($pubkeyA);

        # Extract key string and SSH key version
        if (/ssh-.+ ([\w\d\+\=\/]+) /) {
            $pkversion = 2;
        } elsif (/(\d+ \d+ [\w\d\+\=\/]+)/) {
            $pkversion = 1;
        } else { # Doesn't look like a known public key format, skip it
            next;
        }
        
        # Hash key
        $pubkhash = SHA1_Hash($1);
        close (PUB);

        # Check whether private key is encrypted
        my $pubkeyB = `ssh-keygen -P "" -y -f $privkey_file 2>&1`;
        chomp($pubkeyB);
        if ($pubkeyB =~ /load failed/) {
            $privisenc = 1;
        } else {
            # unencrypted so do additional check to make sure this is a pair
            my $lenB = length $pubkeyB;
            my $pubkeyA = substr ($pubkeyA, 0, $lenB);
            next PKLOOP if ($pubkeyA ne $pubkeyB);
            $privisenc = 0;
        }
        
        my @fields = stat("$pubkey_file");
        ($pkatime, $pkmtime) = ($fields[8], $fields[9]);
        Print_Result "$pubkhash $privisenc $pkversion $pkatime $pkmtime\n";
    }
}

#########################################################################
# Subroutine: Ignore_User($user)
###
#
# Returns 1 if user is to be ignored (i.e. included in @IGNORE_USERS list.
# Returns 0 otherwise.  This is useful for avoiding collection of data from
# system accounts and any users excluded from collection at the command line
#########################################################################
sub Ignore_User
{
    my ($user) = @_;
    foreach (@IGNORE_USERS) {
        return 1 if ($user eq $_);
    }
    
    return 0;
}

#########################################################################
# Subroutine: Encrypt_Results()
###
#
# Encrypt results stored included via Print_Result (i.e. stored in $RESULTS
# variable) using a randomly generated symmetric key.  The symmetric key
# is then encrypted using an RSA public included with the module as 
# 'key.pub'.  Returns the encrypted results.
#########################################################################
sub Encrypt_Results
{
    print "Encrypting results...\n";
    my $key;
    my $enc;

    # Generate a symmetric key
    $key = unpack "H32", `openssl rand 128 2>/dev/null`;

    # RSA encrypt symmetric key
    $enc = "KEY(20050314)\n";

    # Execute openssl command
    open (PIPE, "| openssl rsautl -encrypt -pubin -inkey key.pub | openssl base64 > /tmp/tmp-collect-ssh-$$")
        or die "Can't open RSA pipe: $!";
    print PIPE "$key\n";
    close PIPE;

    # Read back command output
    open (RES, "/tmp/tmp-collect-ssh-$$") 
        or die "Can't open RSA result file: $!";
    foreach (<RES>) {
        $enc .= $_;
    }
    close RES;
    `rm /tmp/tmp-collect-ssh-$$`;

    # 3DES encrypt results with symmetric key
    # We use 3DES rather than AES for compatibilty with older versions 
    # of openssl
    $enc .= "BODY(3DES)\n";

    # Execute openssl command
    open (PIPE, "| openssl enc -des3 -K $key -a -iv 0 >> /tmp/tmp-collect-ssh-$$") 
        or die "Can't open RSA pipe: $!";
    print PIPE "$RESULTS\n";
    close PIPE;

    # Read back command output
    open (RES, "/tmp/tmp-collect-ssh-$$") 
        or die "Can't open RSA result file: $!";
    foreach (<RES>) {
        $enc .= $_;
    }
    close RES;
    `rm /tmp/tmp-collect-ssh-$$`;

    return $enc;
}

#########################################################################
# Subroutine: Save_Results($results, $file)
###
#
# Saves $results (should be encrypted via Encrypt_Results) on disk to $file.
#########################################################################
sub Save_Results
{
    my ($res, $file) = @_;

    if ($file eq "") {
        print "Where do you want to save the results? ";
        $file = <STDIN>;
        chomp($file);
    }

    print "Save results to '$file'\n";
    if (! open (FILE, ">$file")) {
        print STDERR "Failed to open '$file' for writing: $!";
        return -1;
    }

    print FILE $res;
    close (FILE);
    return 0;
}

#########################################################################
# Subroutine: Submit_Results_Mail($results)
###
#
# Submits $results via pipe to local mail program for submission to SSH
# project group at CSAIL
#########################################################################
sub Submit_Results_Mail
{
    if ($DEBUG) {
        print "DEBUG: Successful mail submission\n";
        return 0;
    }

    my ($res) = @_;
    my $recip = 'sshproject@nms.csail.mit.edu';

    if (! open (MAIL, "|mail $recip")) {
        print STDERR "Failed to open pipe to mail: $!";
        return -1;
    }

    print MAIL $res;    
    close (MAIL);
    return 0;
}

#########################################################################
# Subroutine: Submit_Results_HTTP($results)
###
#
# Submits $results by opening a socket to port 80 on SSH project group
# submission server at CSAIL
#########################################################################
sub Submit_Results_HTTP
{
    if ($DEBUG) {
        print "DEBUG: Failed HTTP submission\n";
        return -1;
    }

    my ($res) = @_;
    my $host = 'tennis.lcs.mit.edu';
    my $port = '80';
    my $post;
    my $len;

    print "Submitting results...\n";

    my $protocol = getprotobyname('tcp');
    $host = inet_aton($host) or return -1;

    socket(SOCK, AF_INET, SOCK_STREAM, $protocol) or return -1;
    
    my $dest_addr = sockaddr_in($port,$host);
    connect(SOCK,$dest_addr) or return -1;

    $len = length $res;
    $post = 
        "POST /submit HTTP/1.1\r\n" .
        "Connection: close\r\n" .
        "Content-Type: text/plain\r\n" .
        "Content-Length: $len\r\n\r\n" .
        $res;

    print SOCK $post;
    close SOCK;
    return 0;
}

# Check whether a user's home directory is on a network filesystem
# If it is, resolve, anonymize, and return IP address of the FS host.  
# If not on a network filesystem, return "LOCAL".
# If unresolvable, return "CANTRESOLVE".

#########################################################################
# Subroutine: Homedir_Netfs_IP($home_dir)
###
#
# Checks whether $home_dir is a stored on a local filesystem or on a network
# filesystem via inspecting the output of the df command.  If stored locally,
# returns "LOCAL".  If stord remotely, returned anonymized IP of server.  If
# unable to result IP address of remote server, results "CANTRESOLVE".
#########################################################################
sub Homedir_Netfs_IP
{
    my ($hdir) = @_;

    # Get mount point information for the given user's home directory
    my $dfout = `df $hdir\n`;

    # Does the df output for the user's home directory include
    # a string of format hostname:path, suggesting that the user
    if ($dfout =~ /([\w\d.-]+):[\w\d._\/-]+/) {
        # Resolve hostname to IP address
        my ($name,$aliases,$addrtype,$length,@addrs) = gethostbyname $1;

        return "CANTRESOLVE" if (! defined $addrs[0]);
            
        # Convert resolved IP to dotted quad format & anonymize
        my ($ipa, $ipb, $ipc, $ipd) = unpack('C4', $addrs[0]);
        return Anonymize_IP("$ipa.$ipb.$ipc.$ipd");
    }

    return "LOCAL";
}

#########################################################################
# Subroutine: Read_User($home_dir, $user)
###
#
# Collects all user-specific data (SHA1 hash of username, IP of home
# directory host, results of Read_Known_Hosts, results of 
# Read_Authorized_keys, results of Read_Public_keys).
#########################################################################
sub Read_User
{
    my ($hdir, $user) = @_;
    my $userhash;
    my $hdir_net_ip;
    my $lastlog;

    Print_Result "user information\n";

    $userhash = SHA1_Hash($user);
    Print_Result "$userhash\n";

    chop($lastlog = `last -1 $user`);
    if ($lastlog =~ /\w+\s+[\w\s\/\-\.]+\s+(\w+ \w+ \d+ [\d:]+)\s+[s\-\(]/) {
        Print_Result "last: $1\n";
    } else { # Couldn't parse line, so just send the whole thing
        Print_Result "last(full): $lastlog\n";
    }

    $hdir_net_ip = Homedir_Netfs_IP($hdir);
    Print_Result "$hdir_net_ip\n";

    Print_Result "known_hosts entries\n";
    Read_Known_Hosts("$hdir/.ssh/known_hosts");
    Read_Known_Hosts("$hdir/.ssh/known_hosts2");

    Print_Result "authorized_keys entries\n";
    Read_Authorized_Keys("$hdir/.ssh/authorized_keys");
    Read_Authorized_Keys("$hdir/.ssh/authorized_keys2");

    Print_Result ".pub file entries\n";
    Read_Public_Keys($hdir);
    Print_Result "user information end\n";
}

1;
# EOF
