#!/usr/bin/perl                 -w -ICrypt-Rijndael-0.04/blib/lib -ICrypt-Rijndael-0.04/blib/arch/auto/Crypt/Rijndael -IDigest-SHA1-2.10/blib/lib -IDigest-SHA1-2.10/blib/arch/auto/Digest/SHA1

###########################################################################
# Author List
##############
# Jaeyeon Jung - MIT CSAIL
# Stuart Schechter - Harvard University DEAS
# Will Stockwell - MIT CSAIL
#
# Copyright (c) 2004 Jaeyeon Jung, Stuart Schechter, Will Stockwell
#                    Cambridge, MA, All rights reserved
#
#######################################################################
# Documentation
################
#
#   This script collects data from users' SSH configuration data stored
#   in the ~/.ssh/ for use in a research project being conducted by the
#   NMS group at MIT CSAIL. Please see the README for more details or visit
#   the following URL for additional information:
#
#   http://nms.lcs.mit.edu/projects/ssh/collect-ssh.php
#
#   Command line options:
#

# perl collect-ssh.pl <options>
#   If run as regular user, the script will collect data only for that
#   particular user.  The following options are of interest to those
#   collecting data from a whole system of users.
#
#   Standard options (use a space between option and value):
#   -h           Display this usage message
#   -pf <file>   Specify an alternate path to your system passwd file.
#                Default is '/etc/passwd'. This is useful if your
#                system uses NIS or LDAP which can sometimes cause problems
#                for the script on large systems. Tools to generate passwd
#                file style data from NIS and LDAP can avoid these problems.
#                (see README for more information)
#   -x <user>    Exclude <user> from collection procedure.  Multiple
#                '-x' options may be specified.
#
#   Alternate options to find user home directories (instead of -a or -pf):
#   -u <path>    Collect data from files in the .ssh subdirectory
#                beneath this user home directory path, specified
#                via the <path> value.
#                (e.g. -u /home/johndoe collects data in /home/johndoe/.ssh/)
#   -m <path>    Like -u, but the <path> specifies a master subdirectory
#                under which there are user subdirectories for which the
#                <path>/<user>/.ssh/ directory contains the known_hosts
#                files from which to collect data
#                (e.g. -m /var/home examines files in /var/home/*/.ssh/ and
#                 globbing operations work as well such as -m /var/home/*/*
#                 in the case that user home directories are subdivided
#                 into multiple 'master' subdirectories)
#   -r <path>    Location of superuser home directory. Default is '/root'.
#                This is only necessary if you specify '-u' or '-m' options
#                and have located your superuser home directory somewhere
#                unusual
#
#   Submission method options (do not require superuser):
#   -c http      Submit via tcp connection over port 80 (HTTP)
#                Will not automatically detect proxies, so this will
#                may fail if your firewall blocks outgoing port 80.
#                ('-w' is equivalent)
#   -c mail      Pipe to /bin/mail and send to sshproject\@nms.lcs.mit.edu
#                Works with your existing mail infrastructure to
#                prevent interference from firewalls.  Will fail
#                if host is not able or allowed to send mail.
#                ('-e' is equivalent)
#   -c file:<file>
#                Save to file for manual submission
#                Please email to sshproject\@nms.lcs.mit.edu
#                ('-f <file>' is equivalent)
#   -c prompt    Prompt user for preferred option <default>
#
#   Observing data collection (do not require superuser):
#   -verbose     Show all data collected as it is collected
#                <default unless -s set to http or mail>
#                (-v is equivalent)
#   -quiet       Collect data silently
#                <default if -s set to http or mail>
#                (-q is equivalent)
#
#
#######################
# Terms and conditions
#######################
# This script is to be integrated into the OpenSSL package and licensed
# under the same terms specified for the rest of the OpenSSH package,
# as copied below.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
###########################################################################

use strict;
use CollectSSH;

#########################################################################
# Subroutine: InvalidUsage
###
#
# Print out the invalid usage message to inform the user that the
# command line parameters provided were invalid.  Then exit to script.
#########################################################################
sub InvalidUsage()
{

    print STDERR <<USAGE;

perl collect-ssh.pl <options>
  If run as regular user, the script will collect data only for that
  particular user.  The following options are of interest to those
  collecting data from a whole system of users.

  Standard options (use a space between option and value):
  -h           Display this usage message
  -pf <file>   Specify an alternate path to your system passwd file.
               Default is '/etc/passwd'. This is useful if your
               system uses NIS or LDAP which can sometimes cause problems
               for the script on large systems. Tools to generate passwd
               file style data from NIS and LDAP can avoid these problems.
               (see README for more information)
  -x <user>    Exclude <user> from collection procedure.  Multiple
               '-x' options may be specified.

  Alternate options to find user home directories (instead of -a or -pf):
  -u <path>    Collect data from files in the .ssh subdirectory
               beneath this user home directory path, specified
               via the <path> value.
               (e.g. -u /home/johndoe collects data in /home/johndoe/.ssh/)
  -m <path>    Like -u, but the <path> specifies a master subdirectory
               under which there are user subdirectories for which the
               <path>/<user>/.ssh/ directory contains the known_hosts
               files from which to collect data
               (e.g. -m /var/home examines files in /var/home/*/.ssh/ and
                globbing operations work as well such as -m /var/home/*/*
                in the case that user home directories are subdivided
                into multiple 'master' subdirectories)
  -r <path>    Location of superuser home directory. Default is '/root'.
               This is only necessary if you specify '-u' or '-m' options
               and have located your superuser home directory somewhere
               unusual

  Submission method options (do not require superuser):
  -c http      Submit via tcp connection over port 80 (HTTP)
               Will not automatically detect proxies, so this will
               may fail if your firewall blocks outgoing port 80.
               ('-w' is equivalent)
  -c mail      Pipe to /bin/mail and send to sshproject\@nms.lcs.mit.edu
               Works with your existing mail infrastructure to
               prevent interference from firewalls.  Will fail
               if host is not able or allowed to send mail.
               ('-e' is equivalent)
  -c file:<file>
               Save to file for manual submission
               Please email to sshproject\@nms.lcs.mit.edu
               ('-f <file>' is equivalent)
  -c prompt    Prompt user for preferred option <default>

  Observing data collection (do not require superuser):
  -verbose     Show all data collected as it is collected
               <default unless -s set to http or mail>
               (-v is equivalent)
  -quiet       Collect data silently
               <default if -s set to http or mail>
               (-q is equivalent)

USAGE

    exit 0;
}


##########################################################################
# Main program
###############
#
#                        collect-ssh
#
#
##########################################################################

# A set of root home directories in which home directories are stored
# and that we should traverse for data collection from known_hosts files.
my %MASTER_HOME_PATHS = ();

# A set of the user home directories that we should traverse to
# collect data from known_hosts files.  Each should have a child
# directory named ".ssh" or such.  Keys are user home directories.
# Values are references to arrays [ $uid, $username ].  Will be filled
# with all subdirectories of the members of %MASTER_HOME_PATHS.  Once
# filled, this structure is also used by the collection procedure
# (which is the reason for holding onto UID and username.
my %USER_HOME_PATHS = ();

my $SUBMIT_METHOD = 0;
my $SAVE_FILE = "";
my $PASSWD_FILE;
my $ROOT_HOME_DIR = "/root/";

# Arguments passed as "-u".  Will be screened by "-x" arguments prior to 
# inclusion in %USER_HOME_PATHS
my @ARG_USER_HOME_PATHS;

# Parse command line options
while (my $option = shift @ARGV) {
    if ($option eq "-c") {
        if ($SUBMIT_METHOD) {
            print STDERR "Cannot submit results twice! Attempting first specified method\n";
            InvalidUsage();
        }

        InvalidUsage() unless my $method = shift @ARGV;
        if ($method =~ /[Hh][Tt][Tt][Pp]/) {
            $SUBMIT_METHOD = 1;
            $CollectSSH::VERBOSE = 0;
        } elsif ($method =~ /[Mm][Aa][Ii][Ll]/) {
            $SUBMIT_METHOD = 2;
            $CollectSSH::VERBOSE = 0;
        } elsif ($method =~/[Ff][Ii][Ll][Ee]:([\w\d_-]+)/) {
            $SUBMIT_METHOD = 3;
            $SAVE_FILE = $1;
        } elsif ($method =~/[Pp][Rr][Oo][Mm][Pp][Tt]/) {
            next; # This is the default
        } else {
            print STDERR "Invalid submission method '$method' specified\n";
            InvalidUsage();
        }

    } elsif ($option eq "-debug") {
        $CollectSSH::DEBUG = 1;
    } elsif ($option eq "-e") {
        if ($SUBMIT_METHOD) {
            print STDERR "Cannot submit results twice! Attempting first specified method\n";
            InvalidUsage();
        }

        $SUBMIT_METHOD = 2;
        $CollectSSH::VERBOSE = 0;

    } elsif ($option eq "-f") {
        if ($SUBMIT_METHOD) {
            print STDERR "Cannot submit results twice! Attempting first specified method\n";
            InvalidUsage();
        }

        $SUBMIT_METHOD = 3;
        InvalidUsage() unless $SAVE_FILE = shift @ARGV;

    } elsif ($option eq "-h") {
        InvalidUsage();
    } elsif ($option eq "-m") {
        if (defined $PASSWD_FILE) {
            print STDERR "-pf and -u/-m are mutually exclusive arguments\n\n";
            InvalidUsage();
        } elsif ($> != 0) {
            print STDERR "Superuser privileges required for $option option\n\n";
            InvalidUsage();
        } 

	# User has specified a path to a master home directory
	# in which user home directories are located
	InvalidUsage() unless my $master_home_path = shift @ARGV;

        # Grab as many paths as are specified (i.e. looking for next
        # argument or end of the list)
        while (defined $master_home_path) {
            $master_home_path .= "/" unless
                (substr($master_home_path,length($master_home_path)-1,1) eq "/");
            unless (-d $master_home_path) {
                print STDERR "No such directory: $master_home_path\n\n";
                exit;
            }
            $MASTER_HOME_PATHS{$master_home_path} = 1;

            $master_home_path = shift @ARGV;
            if ((! defined($master_home_path)) ||
                $master_home_path =~ /^\-/) { # Next argument (not another dir)
                unshift (@ARGV, $master_home_path); # Put it back in the list
                last;
            }
        } 

    } elsif ($option eq "-pf") {
        if (scalar(keys %MASTER_HOME_PATHS) != 0) {
            print STDERR "-pf and the -u/-m options are mutually exclusive arguments\n\n";
            InvalidUsage();
        } elsif ($> != 0) {
            print STDERR "Superuser privileges required for $option option\n\n";
            InvalidUsage();
        }

        InvalidUsage() unless $PASSWD_FILE = shift @ARGV;
    } elsif ($option eq "-quiet" || $option eq "-q") {
        $CollectSSH::VERBOSE = 0;

    } elsif ($option eq "-r") {
        if ($> != 0) {
            print STDERR "Superuser privileges required for $option option\n\n";
            InvalidUsage();
        }

        InvalidUsage() unless $ROOT_HOME_DIR = shift @ARGV;
    } elsif ($option eq "-u") {
        if (defined $PASSWD_FILE) {
            print STDERR "-pf and -u/-m are mutually exclusive arguments\n\n";
            InvalidUsage();
        } elsif ($> != 0) {
            print STDERR "Superuser privileges required for $option option\n\n";
            InvalidUsage();
        }

 	# User has specified a path to a user home directory
	InvalidUsage() unless my $user_home_path = shift @ARGV;
        while (defined $user_home_path) {
	    push @ARG_USER_HOME_PATHS, $user_home_path;

            $user_home_path = shift @ARGV;
            if ((! defined($user_home_path)) ||
                $user_home_path =~ /^\-/) { # Next argument
                unshift (@ARGV, $user_home_path); # Put it back in the list
                last;
            }
        }

    } elsif ($option eq "-verbose" || $option eq "-v") {
        $CollectSSH::VERBOSE = 1;

    } elsif ($option eq "-w") {
        if ($SUBMIT_METHOD) {
            print STDERR "Cannot submit results twice! Attempting first specified method\n";
            InvalidUsage();
        }

        $SUBMIT_METHOD = 1;
        $CollectSSH::VERBOSE = 0;

    } elsif ($option eq "-x") {
        if ($> != 0) {
            print STDERR "Superuser privileges required for $option option\n\n";
            InvalidUsage();
        }

        push @CollectSSH::IGNORE_USERS, shift @ARGV;
    } else {
        print STDERR "Invalid option: $option\n";
        InvalidUsage();
    }
}

if ($< != 0) {
    # Not root.  Run against indvidual user's SSH data
    $USER_HOME_PATHS{$ENV{HOME}} = [ (getpwnam($ENV{USER}))[2], $ENV{USER} ];
} else {
    # Root.  Read entries for all users, and collect their SSH data

    if (scalar (keys %MASTER_HOME_PATHS) != 0
        || scalar(@ARG_USER_HOME_PATHS) != 0) {
        # -m options specified.  Expand -m into user home directories
        # Collection script also requires user IDs and user names.
        # Grab these too.

        my ($user, $user_id, $user_home_path);

	foreach $user_home_path (@ARG_USER_HOME_PATHS) {
	    $user_home_path .= "/" unless
		(substr($user_home_path,length($user_home_path)-1,1) eq "/");
	    unless (-d $user_home_path) {
		print STDERR "No such directory: $user_home_path\n\n";
		exit;
	    }

	    $user = (split ("/", $user_home_path))[-1];
	    next if CollectSSH::Ignore_User($user);
	    $user_id = (stat($user_home_path))[4];
	    $USER_HOME_PATHS{$user_home_path} = [ $user_id, $user ];
	}

        foreach my $path (keys %MASTER_HOME_PATHS) {
            opendir (DIR, $path) || die "Failed to open directory $path: $!";
            my @subdirs = grep {!/^\./ && -d "$path/$_" } readdir(DIR);
            close (DIR);

            foreach $user (@subdirs) {
                next if CollectSSH::Ignore_User($user);
                $user_home_path = "$path/$user";
                $user_id = (stat($user_home_path))[4];
                $USER_HOME_PATHS{$user_home_path} = [ $user_id, $user ];
            }
        }

        # Can't forget to include superuser
        $USER_HOME_PATHS{$ROOT_HOME_DIR} = [ 0, "root" ];

    } else {
        # DEFAULT BEHAVIOR is to read a passwd file (either directly
        # from the system or generated via YP, LDAP, etc. commands
        # depending on local system configuration.  This approach is
        # prefered over getpwent style retrieval which will do YP and
        # LDAP work transparently, but in instances the large amount
        # of RPC caused YP systems to time out.  This moves all that
        # work to before the script is executed that should not cause
        # this time out to occur

        $PASSWD_FILE = "/etc/passwd" unless defined $PASSWD_FILE;
        open (PW, "$PASSWD_FILE") || die "failed to open password file: $!";

        foreach (<PW>) {
            my @fields = split ":";
            next if CollectSSH::Ignore_User($fields[0]);
            $USER_HOME_PATHS{$fields[5]} = [ $fields[2], $fields[0] ];
        }

        close (PW);
    }
}

# Perform collection 
CollectSSH::Read_Host();
foreach my $user_hdir (keys %USER_HOME_PATHS) {
    my ($user_id, $user) = @{$USER_HOME_PATHS{$user_hdir}};
    $> = $user_id;
    CollectSSH::Read_User($user_hdir, $user);
    $> = $<;
}

my $enc = CollectSSH::Encrypt_Results();

SUBMIT:
    while (1) {
        my $resp;

        if ($SUBMIT_METHOD > 0 && $SUBMIT_METHOD <= 4) { 
            # $SUBMIT_METHOD is set if a submit method was passed on the
            # command line. It is unset below in the event that the method
            # fails so that the prompt will appear in subsequent loop
            # iterations
            $resp = $SUBMIT_METHOD;
            $SUBMIT_METHOD = 0;
        } else {
            # Either on submission method was specified on the command line or
            # the method failed so it is appropriate to prompt the user for an
            # alternate method

            print <<MSG;
How would you like to submit these data:

[1] Open HTTP connection 
[2] E-mail results 
[3] Save to file for manual e-mail
[4] Do not submit data
MSG

            $resp = <STDIN>;
            chomp($resp);
        }

        if ($resp eq "1") {
            last if (CollectSSH::Submit_Results_HTTP($enc) == 0);
            print "HTTP submission failed. Try another method.\n";
            goto SUBMIT;

        } elsif ($resp eq "2") {
            last if (CollectSSH::Submit_Results_Mail($enc) == 0);
            print "E-mail (via pipe to mail program) submission failed. Try another method.\n";
            goto SUBMIT;

        } elsif ($resp eq "3") {
            last if (CollectSSH::Save_Results($enc, $SAVE_FILE) == 0);
            print "Saving submission to disk failed. Try another method.\n";
            goto SUBMIT;

        } elsif ($resp eq "4") {
            print "Submission cancelled.  Done.\n";
            exit 0;
        }

        print "Bad input. ";
    } # end of while(1) loop

# End perform collection

print "Done!  Thanks for participating!\n";

# EOF
